@@ -8,8 +8,9 @@ mod utils;
 mod worker;
 
 use std::cmp::Ordering;
-use std::fs::File;
+use std::fs::{self, File};
 use std::io::{Read, Write};
+use std::iter::FromIterator;
 use std::path::Path;
 use std::path::PathBuf;
 use std::sync::Arc;
@@ -39,6 +40,7 @@ use crate::worker::{IndexMsg, MainToWorkerMessage, Worker, WorkerToMainMessage};
 use crossbeam::channel::{self, Receiver, Sender};
 use crossbeam::deque::Injector as CrossbeamInjector;
 use glob::Pattern;
+use normalize_line_endings::normalized;
 use rustc_hash::FxHashMap;
 use serde::{Deserialize, Serialize};
 
@@ -169,6 +171,8 @@ pub struct MorselsConfig {
     lang_config: MorselsLanguageConfig,
     #[serde(default)]
     pub indexing_config: MorselsIndexingConfig,
+    #[serde(skip)]
+    pub raw_config: String,
 }
 
 // Separate struct to support serializing for --init option but not output config
@@ -176,6 +180,7 @@ pub struct MorselsConfig {
 struct MorselsIndexingOutputConfig {
     loader_configs: FxHashMap<String, Box<dyn Loader>>,
     pl_names_to_cache: Vec<u32>,
+    num_docs_per_block: u32,
     num_pls_per_dir: u32,
     num_stores_per_dir: u32,
     with_positions: bool,
@@ -184,8 +189,10 @@ struct MorselsIndexingOutputConfig {
 #[derive(Serialize)]
 pub struct MorselsOutputConfig<'a> {
     ver: &'static str,
+    last_doc_id: u32,
     indexing_config: MorselsIndexingOutputConfig,
     lang_config: &'a MorselsLanguageConfig,
+    cache_all_field_stores: bool,
     field_infos: &'a FieldInfos,
 }
 
@@ -195,6 +202,7 @@ impl Default for MorselsConfig {
             indexing_config: MorselsIndexingConfig::default(),
             lang_config: MorselsLanguageConfig::default(),
             fields_config: FieldsConfig::default(),
+            raw_config: "".to_owned(),
         }
     }
 }
@@ -203,7 +211,7 @@ pub struct Indexer {
     indexing_config: Arc<MorselsIndexingConfig>,
     doc_id_counter: u32,
     spimi_counter: u32,
-    pl_names_to_cache: Vec<u32>,
+    cache_all_field_stores: bool,
     field_infos: Arc<FieldInfos>,
     output_folder_path: PathBuf,
     doc_miner: WorkerMiner,
@@ -218,6 +226,7 @@ pub struct Indexer {
     is_dynamic: bool,
     delete_unencountered_external_ids: bool,
     start_doc_id: u32,
+    start_block_number: u32,
     dynamic_index_info: DynamicIndexInfo,
 }
 
@@ -227,9 +236,56 @@ impl Indexer {
         output_folder_path: &Path,
         config: MorselsConfig,
         mut is_dynamic: bool,
+        preserve_output_folder: bool,
         delete_unencountered_external_ids: bool,
     ) -> Indexer {
-        let dynamic_index_info = DynamicIndexInfo::new_from_output_folder(&output_folder_path, &mut is_dynamic);
+        fs::create_dir_all(output_folder_path).expect("could not create output directory!");
+
+        let raw_config_normalised = &String::from_iter(normalized(config.raw_config.chars()));
+
+        let dynamic_index_info = DynamicIndexInfo::new_from_output_folder(
+            &output_folder_path,
+            raw_config_normalised,
+            &mut is_dynamic
+        );
+
+        if !is_dynamic && !preserve_output_folder {
+            if let Ok(read_dir) = fs::read_dir(output_folder_path) {
+                for dir_entry in read_dir {
+                    if let Err(err) = dir_entry {
+                        eprintln!("Failed to clean {}, continuing.", err);
+                        continue;
+                    }
+
+                    let dir_entry = dir_entry.unwrap();
+                    let file_type = dir_entry.file_type();
+                    if let Err(err) = file_type {
+                        eprintln!("Failed to get file type when cleaning output dir {}, continuing.", err);
+                        continue;
+                    }
+
+                    let file_type = file_type.unwrap();
+                    if file_type.is_file() {
+                        if let Err(err) = fs::remove_file(dir_entry.path()) {
+                            eprintln!("{}\nFailed to clean {}, continuing.", err, dir_entry.path().to_string_lossy());
+                        }
+                    } else if file_type.is_dir() {
+                        if let Err(err) = fs::remove_dir_all(dir_entry.path()) {
+                            eprintln!("{}\nFailed to clean directory {}, continuing.", err, dir_entry.path().to_string_lossy());
+                        }
+                    }
+                }
+            } else {
+                eprintln!("Failed to read output dir for cleaning, continuing.");
+            }
+        }
+
+        {
+            File::create(output_folder_path.join("old_morsels_config.json"))
+                .expect("error creating old config file")
+                .write_all(raw_config_normalised.as_bytes())
+                .expect("error writing old config");
+        }
 
         let loaders = config.indexing_config.get_loaders_from_config();
 
@@ -285,7 +341,11 @@ impl Indexer {
         }));
 
         let doc_id_counter = doc_infos.lock().unwrap().doc_lengths.len() as u32;
-        let start_doc_id = doc_id_counter;
+
+        #[cfg(debug_assertions)]
+        println!("previous number of docs {}", doc_id_counter);
+
+        let spimi_counter = doc_id_counter % config.indexing_config.num_docs_per_block;
 
         // Construct worker threads
         let (tx_worker, rx_main): (
@@ -329,7 +389,6 @@ impl Indexer {
                         indexing_config_clone,
                         expected_num_docs_per_thread,
                         num_workers_writing_blocks_clone,
-                        is_dynamic,
                         index_unit_queue,
                     )
                 }),
@@ -341,13 +400,15 @@ impl Indexer {
             indexing_config.with_positions,
             expected_num_docs_per_thread,
             &tokenizer,
+            #[cfg(debug_assertions)]
+            0,
         );
 
-        let indexer = Indexer {
+        let mut indexer = Indexer {
             indexing_config,
             doc_id_counter,
-            spimi_counter: 0,
-            pl_names_to_cache: Vec::new(),
+            spimi_counter,
+            cache_all_field_stores: config.fields_config.cache_all_field_stores,
             field_infos,
             output_folder_path: output_folder_path.to_path_buf(),
             doc_miner,
@@ -361,9 +422,11 @@ impl Indexer {
             lang_config: config.lang_config,
             is_dynamic,
             delete_unencountered_external_ids,
-            start_doc_id,
+            start_doc_id: doc_id_counter,
+            start_block_number: 0,
             dynamic_index_info,
         };
+        indexer.start_block_number = indexer.block_number();
 
         indexer.make_workers_index(num_threads);
 
@@ -374,21 +437,21 @@ impl Indexer {
         match lang_config.lang.as_str() {
             "ascii" => {
                 if let Some(options) = lang_config.options.as_ref() {
-                    Arc::new(ascii::new_with_options(serde_json::from_value(options.clone()).unwrap()))
+                    Arc::new(ascii::new_with_options(serde_json::from_value(options.clone()).unwrap(), false))
                 } else {
                     Arc::new(ascii::Tokenizer::default())
                 }
             }
             "latin" => {
                 if let Some(options) = lang_config.options.as_ref() {
-                    Arc::new(latin::new_with_options(serde_json::from_value(options.clone()).unwrap()))
+                    Arc::new(latin::new_with_options(serde_json::from_value(options.clone()).unwrap(), false))
                 } else {
                     Arc::new(latin::Tokenizer::default())
                 }
             }
             "chinese" => {
                 if let Some(options) = lang_config.options.as_ref() {
-                    Arc::new(chinese::new_with_options(serde_json::from_value(options.clone()).unwrap()))
+                    Arc::new(chinese::new_with_options(serde_json::from_value(options.clone()).unwrap(), false))
                 } else {
                     Arc::new(chinese::Tokenizer::default())
                 }
@@ -400,8 +463,7 @@ impl Indexer {
     }
 
     fn block_number(&self) -> u32 {
-        ((self.doc_id_counter - self.start_doc_id) as f64 / (self.indexing_config.num_docs_per_block as f64)).ceil()
-            as u32
+        ((self.doc_id_counter as f64) / (self.indexing_config.num_docs_per_block as f64)).floor() as u32
     }
 
     fn make_workers_index(&self, n: usize) {
@@ -455,13 +517,20 @@ impl Indexer {
 
                     if self.spimi_counter == self.indexing_config.num_docs_per_block {
                         let mut num_workers_writing_blocks = self.num_workers_writing_blocks.lock().unwrap();
+                        // Minimally need 1 worker to take up this job still
+                        while *num_workers_writing_blocks == self.indexing_config.num_threads {
+                            drop(num_workers_writing_blocks);
+                            std::thread::sleep(std::time::Duration::from_millis(10));
+                            num_workers_writing_blocks = self.num_workers_writing_blocks.lock().unwrap();
+                        }
+
                         let num_active_workers = self.indexing_config.num_threads - *num_workers_writing_blocks;
                         for _i in 0..num_active_workers {
                             self.index_unit_queue.push(IndexMsg::Stop);
                         }
 
-                        let main_thread_block_index_results = self.doc_miner.get_results(1000000);
-                        let block_number = self.block_number();
+                        let main_thread_block_index_results = self.doc_miner.get_results();
+                        let block_number = self.block_number() - 1;
                         self.write_block(
                             main_thread_block_index_results, block_number, false, &mut * num_workers_writing_blocks
                         );
@@ -496,17 +565,20 @@ impl Indexer {
     fn write_morsels_config(&mut self) {
         let serialized = serde_json::to_string(&MorselsOutputConfig {
             ver: MORSELS_VERSION,
+            last_doc_id: self.doc_id_counter,
             indexing_config: MorselsIndexingOutputConfig {
                 loader_configs: std::mem::take(&mut self.loaders)
                     .into_iter()
                     .map(|loader| (loader.get_name(), loader))
                     .collect(),
-                pl_names_to_cache: std::mem::take(&mut self.pl_names_to_cache),
+                pl_names_to_cache: self.dynamic_index_info.pl_names_to_cache.clone(),
+                num_docs_per_block: self.indexing_config.num_docs_per_block,
                 num_pls_per_dir: self.indexing_config.num_pls_per_dir,
                 num_stores_per_dir: self.indexing_config.num_stores_per_dir,
                 with_positions: self.indexing_config.with_positions,
             },
             lang_config: &self.lang_config,
+            cache_all_field_stores: self.cache_all_field_stores,
             field_infos: &self.field_infos,
         })
         .unwrap();
@@ -517,28 +589,38 @@ impl Indexer {
             .unwrap();
     }
 
+    fn is_deletion_only_run(&self) -> bool {
+        self.doc_id_counter == self.start_doc_id
+    }
+
     pub fn finish_writing_docs(mut self, instant: Option<Instant>) {
-        if self.spimi_counter != 0 {
-            #[cfg(debug_assertions)]
-            println!("Writing last spimi block");
+        #[cfg(debug_assertions)]
+        println!("@finish_writing_docs");
 
-            for _i in 0..self.indexing_config.num_threads {
-                self.index_unit_queue.push(IndexMsg::Stop);
-            }
-
-            let mut num_workers_writing_blocks = self.num_workers_writing_blocks.lock().unwrap();
-            let block_number = self.block_number();
-            let main_thread_block_index_results = self.doc_miner.get_results(1000000);
-            self.write_block(
-                main_thread_block_index_results, block_number, true, &mut * num_workers_writing_blocks
-            );
-            self.spimi_counter = 0;
+        for _i in 0..self.indexing_config.num_threads {
+            self.index_unit_queue.push(IndexMsg::Stop);
         }
 
-        self.wait_on_all_workers();
+        let first_block = self.start_block_number;
+        let mut last_block = self.block_number();
+
+        if self.spimi_counter != 0 {
+            #[cfg(debug_assertions)]
+            println!("Writing extra last spimi block");
+
+            let main_thread_block_index_results = self.doc_miner.get_results();
+            self.write_block(main_thread_block_index_results, last_block, true, &mut 0);
+            self.spimi_counter = 0;
+        } else if !self.is_deletion_only_run() {
+            last_block -= 1;
+            self.wait_on_all_workers();
+        }
 
         #[cfg(debug_assertions)]
-        println!("Number of docs: {}", self.doc_id_counter);
+        println!(
+            "Number of docs: {}, First Block {}, Last Block {}",
+            self.doc_id_counter, first_block, last_block,
+        );
 
         if let Some(now) = instant {
             print_time_elapsed(now, "Block indexing done!");
@@ -546,17 +628,19 @@ impl Indexer {
 
         // Merge spimi blocks
         // Go through all blocks at once
-        let num_blocks = self.block_number();
+        let num_blocks = last_block - first_block + 1;
         if self.is_dynamic {
             if self.delete_unencountered_external_ids {
                 self.dynamic_index_info.delete_unencountered_external_ids();
             }
 
             spimireader::dynamic::modify_blocks(
+                self.is_deletion_only_run(),
                 self.doc_id_counter,
                 num_blocks,
+                first_block,
+                last_block,
                 &self.indexing_config,
-                &mut self.pl_names_to_cache,
                 std::mem::take(&mut self.doc_infos),
                 &self.tx_main,
                 &self.output_folder_path,
@@ -566,8 +650,9 @@ impl Indexer {
             spimireader::full::merge_blocks(
                 self.doc_id_counter,
                 num_blocks,
+                first_block,
+                last_block,
                 &self.indexing_config,
-                &mut self.pl_names_to_cache,
                 std::mem::take(&mut self.doc_infos),
                 &self.tx_main,
                 &self.output_folder_path,
@@ -579,7 +664,9 @@ impl Indexer {
 
         self.dynamic_index_info.write(&self.output_folder_path, self.doc_id_counter);
 
-        spimireader::common::cleanup_blocks(num_blocks, &self.output_folder_path);
+        if !self.is_deletion_only_run() {
+            spimireader::common::cleanup_blocks(first_block, last_block, &self.output_folder_path);
+        }
 
         if let Some(now) = instant {
             print_time_elapsed(now, "Blocks merged!");
