@@ -15,6 +15,12 @@ struct CliArgs {
     source_folder_path: PathBuf,
     #[structopt(parse(from_os_str))]
     output_folder_path: PathBuf,
+    #[structopt(
+        short,
+        long,
+        help = "Preserves the output directory contents, overwriting them as necessary if running a full reindex"
+    )]
+    preserve_output_folder: bool,
     #[structopt(short, long, parse(from_os_str))]
     config_file_path: Option<PathBuf>,
     #[structopt(short, long, help = "Initialise the configuration file in the source folder")]
@@ -82,21 +88,32 @@ fn main() {
         config_file_path.to_str().unwrap(),
     );
 
+    if args.init {
+        morsels_indexer::Indexer::write_morsels_source_config(MorselsConfig::default(), &config_file_path);
+        return;
+    }
+
     let config: MorselsConfig = if config_file_path.exists() && config_file_path.is_file() {
-        let config_raw = std::fs::read_to_string(&config_file_path).unwrap();
-        serde_json::from_str(&config_raw).expect("morsels_config.json does not match schema!")
+        let raw_config = std::fs::read_to_string(&config_file_path).unwrap();
+        let mut config: MorselsConfig = serde_json::from_str(&raw_config).expect("morsels_config.json does not match schema!");
+        config.raw_config = raw_config;
+        config
+    } else if args.config_file_path.is_some() {
+        eprintln!("Specified configuration file {} not found!", config_file_path.to_str().unwrap());
+        return;
     } else {
         MorselsConfig::default()
     };
 
-    if args.init {
-        morsels_indexer::Indexer::write_morsels_source_config(config, &config_file_path);
-        return;
-    }
-
     let exclude_patterns = config.indexing_config.get_excludes_from_config();
 
-    let mut indexer = morsels_indexer::Indexer::new(&output_folder_path, config, args.dynamic, true);
+    let mut indexer = morsels_indexer::Indexer::new(
+        &output_folder_path,
+        config,
+        args.dynamic,
+        args.preserve_output_folder,
+        true,
+    );
 
     let now = if args.perf { Some(Instant::now()) } else { None };
 
